# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: MIT


import pathlib

import pkg_resources


def get_package_version():
    root = pathlib.Path(__file__) / ".."
    root = root.resolve()

    return (root / "VERSION").read_text(encoding="utf-8").rstrip()


def get_runtime_version(pkg):
    """
    `fc_common` including `VERSION` maybe upgraded by other caller packages,
    so internal package version is the most reliable item,
    fallback to `VERSION` only when in docker release
    """

    try:
        version = pkg_resources.get_distribution(pkg).version
    except Exception:  # pylint: disable=broad-except
        version = get_package_version()

    return version



# -*- coding: utf-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: MIT


import logging
import os
from contextlib import suppress

# pylint: disable=too-few-public-methods


class Logger:
    @staticmethod
    def init(
        logger_name, log_name="run.log", log_type="both", log_file_permission=None
    ):
        log_path = os.path.expanduser(os.environ.get("FC_LOG_PATH", "~/.fc/log"))
        log_file = os.path.join(log_path, log_name)

        if not os.path.exists(log_path):
            os.makedirs(log_path)

        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.INFO)

        s_handler = logging.StreamHandler()
        f_handler = logging.FileHandler(log_file)
        if log_file_permission:
            with suppress(Exception):
                os.chmod(log_file, log_file_permission)

        s_format = logging.Formatter(
            fmt="%(asctime)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        f_format = logging.Formatter(
            fmt="%(asctime)s %(message)s - [%(name)s:%(filename)s:%(lineno)d]",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

        s_handler.setFormatter(s_format)
        f_handler.setFormatter(f_format)

        if log_type == "both":
            logger.addHandler(s_handler)
            logger.addHandler(f_handler)
        elif log_type == "stream_only":
            logger.addHandler(s_handler)
        elif log_type == "file_only":
            logger.addHandler(f_handler)



# -*- coding: utf-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
import logging
from contextlib import suppress
from string import Template

import flatdict
from aiohttp import web

from fc_server.core import AsyncRunMixin
from fc_server.core.config import Config


class ApiSvr(AsyncRunMixin):
    """
    Rest api server
    """

    def __init__(self, context):
        self.context = context
        self.external_info_tool = Config.api_server.get("external_info_tool", "")
        self.logger = logging.getLogger("fc_server")

    @staticmethod
    def friendly_status(status):
        with suppress(Exception):
            return Config.frameworks_config[status]["friendly_status"]
        return status

    async def resource_status(self, request):
        return await self.fetch_resource_status(request)

    async def verbose_resource_status(self, request):
        return await self.fetch_resource_status(request, True)

    async def fetch_resource_status(
        self, request, verbose=False
    ):  # pylint: disable=too-many-branches, too-many-locals, too-many-statements
        # get all labgrid managed resources
        labgrid_managed_resources = []
        comments = {}
        for framework in self.context.framework_instances:
            if framework.__module__.split(".")[-1] == "labgrid":
                labgrid_managed_resources = framework.managed_resources
                if verbose:
                    comments = await framework.labgrid_get_comments()
                break

        res = request.match_info.get("res", "")

        resources_info = []
        if res:
            if res in Config.managed_resources:
                item = []
                item.append(res)
                item.append(Config.managed_resources_farm_types.get(res, ""))
                item.append(
                    ApiSvr.friendly_status(
                        self.context.managed_resources_status.get(res, "")
                    )
                )
                if res not in labgrid_managed_resources:
                    item.append("non-debuggable")
                else:
                    item.append("")

                if verbose:
                    try:
                        item.append(comments[res])
                    except KeyError:
                        item.append("")

                if self.external_info_tool:
                    # fetch external resource info if needed
                    fc_resource = res
                    fc_farm_type = Config.managed_resources_farm_types.get(res, "")
                    template = Template(self.external_info_tool)
                    tool_command = template.substitute(
                        {
                            "fc_resource": fc_resource,
                            "fc_farm_type": fc_farm_type,
                            "fc_peripheral_info": "",
                        }
                    )
                    ret, info, _ = await self._run_cmd(tool_command)

                    if ret == 0:
                        item.append(info)
                    else:
                        item.append("NA")
                else:
                    item.append([])

                resources_info.append(item)
        else:
            params = request.rel_url.query
            farm_type = params.get("farmtype", "")
            device_type = params.get("devicetype", "")
            peripheral_info = params.get("peripheralinfo", "")

            if device_type and farm_type:
                scope = Config.raw_managed_resources.get(farm_type, {}).get(
                    device_type, []
                )
            elif device_type:
                scope = []
                for raw_managed_resource in Config.raw_managed_resources.values():
                    scope += raw_managed_resource.get(device_type, [])
            elif farm_type:
                scope = flatdict.FlatterDict(
                    Config.raw_managed_resources.get(farm_type, {})
                ).values()
            else:
                scope = ["all"]

            tool_command_list = []
            for resource, status in self.context.managed_resources_status.items():
                if scope != ["all"] and resource not in scope:
                    continue

                item = []
                item.append(resource)
                item.append(Config.managed_resources_farm_types.get(resource, ""))
                item.append(ApiSvr.friendly_status(status))
                if resource not in labgrid_managed_resources:
                    item.append("non-debuggable")
                else:
                    item.append("")

                if verbose:
                    try:
                        item.append(comments[resource])
                    except KeyError:
                        item.append("")

                resources_info.append(item)

                # fetch external resource info if needed
                if self.external_info_tool and device_type:
                    fc_resource = resource
                    fc_farm_type = Config.managed_resources_farm_types.get(resource, "")
                    template = Template(self.external_info_tool)
                    tool_command = template.substitute(
                        {
                            "fc_resource": fc_resource,
                            "fc_farm_type": fc_farm_type,
                            "fc_peripheral_info": peripheral_info,
                        }
                    )
                    tool_command_list.append(self._run_cmd(tool_command))

            fc_not_match_list = []
            if self.external_info_tool and device_type:
                external_info_list = await asyncio.gather(*tool_command_list)
                for index, value in enumerate(external_info_list):
                    if value[0] == 0:
                        resources_info[index].append(value[1])
                        if value[1].strip() == "FC_NOT_MATCH":
                            fc_not_match_list.append(index)
                    else:
                        resources_info[index].append("NA")
            elif (
                device_type
            ):  # for fc instance which doesn't configure external_info_tool
                for index, resource in enumerate(resources_info):
                    resource.append([])
                    if (
                        peripheral_info
                    ):  # external_info_tool not specified instance defintly not match anything
                        fc_not_match_list.append(index)

            resources_info = [
                item
                for index, item in enumerate(resources_info)
                if index not in fc_not_match_list
            ]

        return web.json_response(resources_info)

    @staticmethod
    async def pong(_):
        return web.Response(text="pong")

    async def booking(self, _):
        cmd = "labgrid-client who | grep -v fc"
        _, bookings_text, _ = await self._run_cmd(cmd)

        bookings_text_list = []
        header = True
        anchor = -1
        for booking_text in bookings_text.splitlines():
            if header:
                anchor = booking_text.find("Place")
                bookings_text_list.append(booking_text)
                header = False
                continue

            if booking_text[anchor:].split(" ")[0] in Config.managed_resources:
                bookings_text_list.append(booking_text)

        return web.Response(text="\n".join(bookings_text_list))

    async def start(self, **svr_args):
        app = web.Application()

        access_logger = logging.getLogger("aiohttp.access")
        access_logger.setLevel(logging.INFO)
        handler = logging.StreamHandler()
        s_format = logging.Formatter(
            fmt="%(asctime)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        handler.setFormatter(s_format)
        access_logger.addHandler(handler)

        app.add_routes([web.get("/ping", self.pong)])
        app.add_routes([web.get("/booking", self.booking)])
        app.add_routes([web.get("/resource", self.resource_status)])
        app.add_routes([web.get("/resource/{res}", self.resource_status)])
        app.add_routes([web.get("/verbose_resource", self.verbose_resource_status)])
        app.add_routes(
            [web.get("/verbose_resource/{res}", self.verbose_resource_status)]
        )

        app_runner = web.AppRunner(app)
        await app_runner.setup()

        api_interface = "0.0.0.0"
        api_port = svr_args["port"]
        loop = asyncio.get_event_loop()
        await loop.create_server(app_runner.server, api_interface, api_port)
        self.logger.info("Api Server ready at http://%s:%d.", api_interface, api_port)



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


from unittest.mock import MagicMock

import pytest

from fc_server.plugins.labgrid import Plugin


@pytest.fixture(name="plugin")
def labgrid_plugin():
    config = {
        "lg_crossbar": "ws://$labgrid_crossbar_ip:20408/ws",
        "priority": 2,
        "seize": False,
    }
    return Plugin(config)


# pylint: disable=protected-access
class TestPluginLabgrid:
    @pytest.mark.asyncio
    async def test_force_kick_off(self, asyncio_patch, mocker, plugin):
        token = "7GN5HFQOEU"
        asyncio_patch(
            mocker, "fc_server.plugins.labgrid.Plugin.labgrid_get_place_token", token
        )

        ret = {
            "Reservation '7GN5HFQOEU'": {
                "owner": "foo/bar",
                "token": "7GN5HFQOEU",
                "state": "acquired",
                "prio": "100.0",
                "filters": {
                    "main": "name=$resource1",
                },
                "allocations": {
                    "main": "$resource1",
                },
                "created": "2023-03-28 10:27:14.881492",
                "timeout": "2023-03-28 11:20:01.893404",
            }
        }
        asyncio_patch(
            mocker, "fc_server.plugins.labgrid.Plugin.labgrid_get_reservations", ret
        )

        mocker_labgrid_cancel_reservation = asyncio_patch(
            mocker,
            "fc_server.plugins.labgrid.Plugin.labgrid_cancel_reservation",
            MagicMock(),
        )
        mocker_labgrid_release_place = asyncio_patch(
            mocker,
            "fc_server.plugins.labgrid.Plugin.labgrid_release_place",
            MagicMock(),
        )

        await plugin.force_kick_off("$resource1")
        mocker_labgrid_cancel_reservation.assert_called()
        mocker_labgrid_release_place.assert_called()

    @pytest.mark.asyncio
    async def test_seize_resource(self, asyncio_patch, mocker, coordinator, plugin):
        mocker_coordinate_resources = asyncio_patch(
            mocker,
            "fc_server.core.coordinator.Coordinator.coordinate_resources",
            MagicMock(),
        )
        await plugin._Plugin__seize_resource(coordinator, "foo", ["$resource1"])

        mocker_coordinate_resources.assert_called()

    @pytest.mark.asyncio
    async def test_schedule(self, asyncio_patch, mocker, plugin, coordinator):
        ret = {
            "Reservation '83UF223356'": {
                "owner": "fc/fc",
                "token": "83UF223356",
                "state": "acquired",
                "prio": "100.0",
                "filters": {
                    "main": "name=$resource1",
                },
                "allocations": {
                    "main": "$resource1",
                },
                "created": "2023-03-28 09:27:14.881492",
                "timeout": "2023-03-28 10:20:01.893404",
            },
            "Reservation '7GN5HFQOEU'": {
                "owner": "foo/bar",
                "token": "7GN5HFQOEU",
                "state": "waiting",
                "prio": "0",
                "filters": {
                    "main": "name=$resource1",
                },
                "allocations": {
                    "main": "$resource1",
                },
                "created": "2023-03-28 10:27:14.881492",
                "timeout": "2023-03-28 11:20:01.893404",
            },
        }
        asyncio_patch(
            mocker, "fc_server.plugins.labgrid.Plugin.labgrid_get_reservations", ret
        )

        ret = (True, True)
        asyncio_patch(
            mocker, "fc_server.plugins.lava.Plugin.default_framework_disconnect", ret
        )

        asyncio_patch(
            mocker,
            "fc_server.plugins.labgrid.Plugin._Plugin__labgrid_guard_reservation",
            MagicMock(),
        )
        asyncio_patch(
            mocker,
            "fc_server.plugins.labgrid.Plugin.labgrid_cancel_reservation",
            MagicMock(),
        )
        asyncio_patch(
            mocker,
            "fc_server.plugins.labgrid.Plugin.labgrid_release_place",
            MagicMock(),
        )

        async def mocker_labgrid_system_reservation():
            pass

        mocker_labgrid_system_reservation = mocker.patch(
            "fc_server.plugins.labgrid.Plugin._Plugin__labgrid_system_reservation",
            return_value=mocker_labgrid_system_reservation(),
        )

        await plugin.schedule(coordinator)
        mocker_labgrid_system_reservation.assert_called()



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


import logging
import multiprocessing
import os
import signal
import time

import prctl

from fc_common.etcd import Etcd
from fc_common.logger import Logger


class ServerDaemon:
    def __init__(self, daemon_paras):
        self.daemon_paras = daemon_paras
        self.server_daemon = None
        self.logger = logging.getLogger("fc_server_daemon")

    def __action(self):
        prctl.set_pdeathsig(signal.SIGHUP)  # pylint: disable=no-member

        Logger.init("fc_server_daemon", "fc_server_daemon.log", log_type="file_only")

        self.logger.info("Start fc-server-daemon")
        self.logger.info(self.daemon_paras)

        etcd = Etcd(self.daemon_paras["etcd"])

        # pylint: disable=protected-access
        self.logger.info("Current endpoint: %s", etcd()._current_endpoint_label)

        etcd.put(
            f"/instances/{self.daemon_paras['instance_name']}/fc",
            self.daemon_paras["fc"],
        )
        etcd.put(
            f"/instances/{self.daemon_paras['instance_name']}/lg",
            self.daemon_paras["lg"],
        )
        for device in self.daemon_paras["devices"]:
            etcd.put(f"/devices/{device}", self.daemon_paras["instance_name"])

        ttl = 60
        while True:
            try:
                self.logger.info("  - try to lock instance")
                with etcd().lock(
                    f"instances/{self.daemon_paras['instance_name']}", ttl
                ) as lock:
                    self.logger.info("  - instance locked successfully")
                    while True:
                        time.sleep(ttl / 4)
                        lock.refresh()
            except Exception as lock_exec:  # pylint: disable=broad-except
                self.logger.debug(lock_exec)

            time.sleep(30)

    def handler(self, *_):
        if self.server_daemon:
            os.kill(self.server_daemon.pid, signal.SIGINT)

    def run(self):
        self.server_daemon = multiprocessing.Process(target=self.__action, daemon=True)
        self.server_daemon.start()

        signal.signal(signal.SIGTERM, self.handler)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


import json
import logging
import os
import signal
import socket
from threading import Lock

import daemon
from daemon import pidfile

from fc_common.config import Config
from fc_common.etcd import Etcd
from fc_common.logger import Logger


class ClientDaemon:
    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self.logger = logging.getLogger("fc_client_daemon")

        etcd_url = Config.load_cfg().get("etcd")
        self.logger.info("Etcd url: %s", etcd_url)
        self.etcd = Etcd(etcd_url)
        self.logger.info("Current endpoint: %s", self.etcd()._current_endpoint_label)

        self.locks_prefix = "/locks/instances/"
        self.locks_prefix_len = len(self.locks_prefix)

        self.devices_prefix = "/devices/"
        self.devices_prefix_len = len(self.devices_prefix)

        self.instance_data = {}
        self.device_data = {}

        self.lock = Lock()

        self.ipc_server_address = "\0/tmp/fc/fc_client_daemon.sock"

    def watch_locks_callback(self, event):
        try:
            for per_event in event.events:
                if isinstance(per_event, Etcd.DeleteEvent):
                    self.logger.info("Delete event:")
                    self.logger.info(per_event.key)

                    instance_name = per_event.key[self.locks_prefix_len :].decode(
                        "utf-8"
                    )
                    with self.lock:
                        self.instance_data.pop(instance_name)

                elif isinstance(per_event, Etcd.PutEvent):
                    self.logger.info("Put event:")
                    self.logger.info(per_event.key)

                    instance_name = per_event.key[self.locks_prefix_len :].decode(
                        "utf-8"
                    )
                    fc_addr = self.etcd.get("/instances/" + instance_name + "/fc")[
                        0
                    ].decode("utf-8")
                    lg_addr = self.etcd.get("/instances/" + instance_name + "/lg")[
                        0
                    ].decode("utf-8")

                    self.logger.info(fc_addr)
                    self.logger.info(lg_addr)

                    with self.lock:
                        self.instance_data.setdefault(instance_name, {})
                        self.instance_data[instance_name]["fc"] = fc_addr
                        self.instance_data[instance_name]["lg"] = lg_addr
        except Exception as locks_cb_exec:  # pylint: disable=broad-except
            self.logger.info(locks_cb_exec)
            # Exit due to issue of https://github.com/kragniz/python-etcd3/issues/1026
            os.kill(os.getpid(), signal.SIGINT)

    def watch_devices_callback(self, event):
        try:
            for per_event in event.events:
                if isinstance(per_event, Etcd.PutEvent):
                    self.logger.info("put device event")

                    device_value = per_event.value.decode("utf-8")
                    device_name = per_event.key[self.devices_prefix_len :].decode(
                        "utf-8"
                    )

                    self.logger.info(device_name)
                    self.logger.info(device_value)

                    with self.lock:
                        self.device_data[device_name] = device_value
        except Exception as devices_cb_exec:  # pylint: disable=broad-except
            self.logger.info(devices_cb_exec)
            # Exit due to issue of https://github.com/kragniz/python-etcd3/issues/1026
            os.kill(os.getpid(), signal.SIGINT)

    def start_data_channel(self):
        self.logger.info("Start data channel.")

        instances = self.etcd.get_prefix(self.locks_prefix)
        for instance in instances:
            instance_name = instance[1].key[self.locks_prefix_len :].decode("utf-8")
            fc_addr = self.etcd.get("/instances/" + instance_name + "/fc")[0].decode(
                "utf-8"
            )
            lg_addr = self.etcd.get("/instances/" + instance_name + "/lg")[0].decode(
                "utf-8"
            )
            self.logger.info(instance)
            self.logger.info(fc_addr)
            self.logger.info(lg_addr)

            self.instance_data.setdefault(instance_name, {})
            self.instance_data[instance_name]["fc"] = fc_addr
            self.instance_data[instance_name]["lg"] = lg_addr

        devices = self.etcd.get_prefix(self.devices_prefix)
        for device in devices:
            self.device_data[
                device[1].key[self.devices_prefix_len :].decode("utf-8")
            ] = device[0].decode("utf-8")

        self.etcd.add_watch_prefix_callback(
            self.locks_prefix, self.watch_locks_callback
        )
        self.etcd.add_watch_prefix_callback(
            self.devices_prefix, self.watch_devices_callback
        )

    def start_ipc_server(self):
        self.logger.info("Start ipc server.")

        server = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        server.bind(self.ipc_server_address)
        server.listen()

        while True:
            client_socket, _ = server.accept()
            data = client_socket.recv(1024)

            self.logger.info(data)

            msg = json.loads(data.decode("utf-8"))
            with self.lock:
                if msg["msg_type"] == "require_info":
                    if msg["para"] == "all":
                        self.logger.info(self.instance_data)
                        client_socket.send(
                            json.dumps(self.instance_data).encode("utf-8")
                        )
                    else:
                        try:
                            logger.info(self.device_data[msg["para"]])
                            return_data = self.instance_data[
                                self.device_data[msg["para"]]
                            ]
                        except Exception as ipc_exec:  # pylint: disable=broad-except
                            return_data = {}
                            self.logger.info(ipc_exec)
                        client_socket.send(json.dumps(return_data).encode("utf-8"))
                elif msg["msg_type"] == "daemon_stop":
                    client_socket.send("ok")
                    os.kill(os.getpid(), signal.SIGINT)


if __name__ == "__main__":
    TMP_FC_PATH = "/tmp/fc"
    if not os.path.exists(TMP_FC_PATH):
        os.makedirs(TMP_FC_PATH)
        os.chmod(TMP_FC_PATH, 0o777)

    os.environ["FC_LOG_PATH"] = TMP_FC_PATH
    Logger.init(
        "fc_client_daemon",
        "fc_client_daemon.log",
        log_type="file_only",
        log_file_permission=0o777,
    )

    logger = logging.getLogger("fc_client_daemon")
    logger.info("Start fc-client-daemon")

    logger_io = [handler.stream for handler in logger.handlers]
    with daemon.DaemonContext(
        umask=0o002,
        pidfile=pidfile.TimeoutPIDLockFile("/tmp/fc/fc_client_daemon.pid"),
        files_preserve=logger_io,
    ) as context:
        try:
            client_daemon = ClientDaemon()
            client_daemon.start_data_channel()
            client_daemon.start_ipc_server()
        except Exception as daemon_exec:  # pylint: disable=broad-except
            logger.info(daemon_exec)



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT

from unittest.mock import MagicMock, call

from fc_client.client import Client


class TestClient:
    def test_status(self, mocker, capsys):
        class Args:
            resource = "imx8mm-evk-sh11"
            farm_type = "bsp"
            device_type = "imx8mm-evk"
            verbose = 0

        mocker.patch(
            "fc_client.client.Client.fetch_metadata",
            return_value={},
        )

        Client.status(Args)

        output = capsys.readouterr()[0]

        assert output in (
            """+----------+------+--------+------+
| Resource | Farm | Status | Note |
+----------+------+--------+------+
+----------+------+--------+------+
""",
        )

    def test_lock(self, mocker):
        class Args:
            resource = "imx8mm-evk-sh11"

        class Output:
            def __init__(self):
                self.text = '[["imx8mm-evk-sh11", "bsp", "idle", ""]]'

        mocker.patch(
            "subprocess.getstatusoutput",
            return_value=(
                0,
                "  acquired: fc/fc",
            ),
        )
        mocker.patch(
            "fc_client.client.Client.fetch_metadata",
            return_value={"fc": "foo", "lg": "bar"},
        )
        mocker.patch("requests.get", return_value=Output())

        reserve_cmd = mocker.patch("subprocess.Popen", return_value=MagicMock())
        lock_cmd = mocker.patch("subprocess.call", return_value=MagicMock())
        Client.lock(Args)

        reserve_cmd.assert_called_with(
            ["labgrid-client", "reserve", "--wait", "name=imx8mm-evk-sh11"],
            stderr=-1,
            stdout=-1,
        )
        lock_cmd.assert_called_with(["labgrid-client", "-p", "imx8mm-evk-sh11", "lock"])

    def test_unlock(self, mocker):
        class Args:
            resource = "imx8mm-evk-sh11"

        mocker.patch(
            "fc_client.client.Client.fetch_metadata",
            return_value={"fc": "foo", "lg": "bar"},
        )

        mocker.patch("os.environ.get", return_value="test")
        mocker.patch(
            "subprocess.getstatusoutput",
            return_value=(
                0,
                """Place 'imx8mm-evk-sh11':
  matches:
    */imx8mm-evk-sh11/*
  acquired: test/test
  acquired resources:
  created: 2022-03-03 10:38:04.453874
  changed: 2023-03-28 16:17:05.881561
  reservation: AU7AKIDBHT
""",
            ),
        )
        mocker.patch(
            "subprocess.check_output",
            return_value="""Reservation 'AU7AKIDBHT':
  owner: test/test
  token: AU7AKIDBHT
  state: acquired
  filters:
    main: name=imx8mm-evk-sh11
  allocations:
    main: imx8mm-evk-sh11
  created: 2023-03-28 16:40:36.097286
  timeout: 2023-03-28 16:46:18.902555
""",
        )
        calls = [
            call(["labgrid-client", "cancel-reservation", "AU7AKIDBHT"]),
            call(["labgrid-client", "-p", "imx8mm-evk-sh11", "unlock"]),
        ]
        labgird_free = mocker.patch("subprocess.call", return_value=MagicMock())

        Client.unlock(Args)

        labgird_free.assert_has_calls(calls)



# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: MIT


import logging
import traceback

import yaml

from fc_common import which
from fc_server.core import AsyncRunMixin


class Labgrid(AsyncRunMixin):
    @which(
        "labgrid-client",
        "Use 'pip3 install labgrid-client' to install labgrid software please.",
    )
    def __init__(self):
        self.logger = logging.getLogger("fc_server")

    async def labgrid_get_places(self):
        cmd = "labgrid-client -v p"
        _, places, _ = await self._run_cmd(cmd)
        place_line = places.splitlines()
        return [line.split("'")[1] for line in place_line if line.find("Place") >= 0]

    async def labgrid_get_comments(self):
        comments = {}
        cmd = "labgrid-client -v p"
        _, places, _ = await self._run_cmd(cmd)
        place_line = places.splitlines()
        try_to_parse_comment = False
        for line in place_line:
            if line.find("Place") >= 0:
                place = line.split("'")[1]
                try_to_parse_comment = True
                continue
            if try_to_parse_comment:
                if line.find("comment: ") > 0:
                    comments[place] = line.split("comment: ")[1]
                try_to_parse_comment = False
        return comments

    async def labgrid_get_reservations(self):
        cmd = "labgrid-client reservations"
        _, reservations_text, _ = await self._run_cmd(cmd)
        try:  # pylint: disable=too-many-nested-blocks
            reservations = yaml.load(reservations_text, Loader=yaml.FullLoader)
        except yaml.YAMLError:
            self.logger.error(traceback.format_exc())
            return

        return reservations

    async def labgrid_create_reservation(
        self, place, priority=None, wait=False, shell=False, timeout=None
    ):
        cmd = f"labgrid-client reserve name={place}"
        if timeout:
            cmd = f"timeout {timeout} " + cmd
        if shell:
            cmd += " --shell"
        if wait:
            cmd += " --wait"
        if priority:
            cmd += f" --prio {priority}"
        ret_cmd = await self._run_cmd(cmd)

        if shell and ret_cmd[0] == 0:
            token_string = ret_cmd[1].split("export LG_TOKEN=")
            reservation = None
            if len(token_string) == 2:
                reservation = token_string[1]
            return ret_cmd, reservation

        return ret_cmd

    async def labgrid_cancel_reservation(self, reservation, quiet=False):
        cmd = f"labgrid-client cancel-reservation {reservation}"
        if quiet:
            cmd += " > /dev/null 2>&1"
        await self._run_cmd(cmd)

    async def labgrid_acquire_place(self, place):
        cmd = f"labgrid-client -p {place} acquire"
        return await self._run_cmd(cmd)

    async def labgrid_release_place(self, place, force=False, quiet=False):
        cmd = f"labgrid-client -p {place} release"
        if force:
            cmd += " -k"
        if quiet:
            cmd += " > /dev/null 2>&1"
        await self._run_cmd(cmd)

    async def labgrid_get_place_token(self, place):
        cmd = f"labgrid-client -p {place} show"
        _, place_info_text, _ = await self._run_cmd(cmd)

        token = ""
        place_info_lines = place_info_text.splitlines()
        for line in place_info_lines:
            if line.find("reservation") >= 0:
                token = line.split(":")[-1].strip()
                break
        return token

    async def labgrid_get_place_owner(self, place):
        cmd = f"labgrid-client -p {place} show"
        _, place_info_text, _ = await self._run_cmd(cmd)

        owner = ""
        place_info_lines = place_info_text.splitlines()
        for line in place_info_lines:
            if line.find("acquired:") >= 0:
                owner = line.split(":")[-1].strip()
                break
        return owner



# -*- coding: utf-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
import logging
import os

from fc_server.core.decorators import (
    check_priority_scheduler,
    check_seize_strategy,
    safe_cache,
)
from fc_server.core.plugin import FCPlugin
from fc_server.plugins.utils.labgrid import Labgrid


class Plugin(FCPlugin, Labgrid):
    """
    Plugin for [labgrid framework](https://github.com/labgrid-project/labgrid)
    """

    def __init__(self, frameworks_config):
        super().__init__()
        self.schedule_interval = 2  # poll labgrid reserve queues every 2 seconds

        os.environ["LG_CROSSBAR"] = frameworks_config["lg_crossbar"]
        os.environ["LG_HOSTNAME"] = "fc"
        os.environ["LG_USERNAME"] = "fc"

        self.managed_resources = None

        self.seize_cache = {}  # cache to avoid busy seize

        self.logger = logging.getLogger("fc_server")

    @safe_cache
    def __update_cache(self, cache_name, job_id, value):
        self.__dict__[cache_name][job_id] += value

    async def __labgrid_guard_reservation(self, resource):
        self.logger.info("* [start] inject guard reservation for %s", resource)
        await self.labgrid_create_reservation(resource, priority=-100)
        self.logger.info("* [done] inject guard reservation for %s", resource)

    async def __labgrid_system_reservation(self, driver, resource):
        self.logger.info("* [start] inject fc reservation for %s", resource)
        await self.labgrid_create_reservation(resource, priority=100, wait=True)
        await self.labgrid_acquire_place(resource)
        self.logger.info("* [done] inject fc reservation for %s", resource)

        await driver.return_resource(resource)

    async def __labgrid_init(self, driver, resource):
        """
        Let FC take over by inject special reservation
        """
        system_reservation_found = False

        reservations = await self.labgrid_get_reservations()
        if reservations:
            for _, v in reservations.items():  # pylint: disable=invalid-name
                if (
                    v["filters"]["main"] == f"name={resource}"
                    and v["owner"] == "fc/fc"
                    and v["prio"] == 100.0
                ):
                    if v["state"] == "acquired":
                        # do nothing if system reservation already there
                        self.logger.info("- %s system reservation exist", resource)
                        system_reservation_found = True
                    else:
                        await self.labgrid_cancel_reservation(v["token"])

        if not system_reservation_found:
            # add system reservation for bare lock which previously not in FC control
            # or system reservation expired
            _, reservation = await self.labgrid_create_reservation(
                resource, priority=100, shell=True
            )

            ret, _, _ = await self.labgrid_acquire_place(resource)
            if ret != 0:
                # either this resource directly be locked before FC take control,
                # or fall into schedule gap when FC restart
                if reservation:
                    await self.labgrid_cancel_reservation(reservation)
                asyncio.create_task(self.__labgrid_system_reservation(driver, resource))
                self.logger.info("- %s system reservation scheduled", resource)
            else:
                self.logger.info("- %s system reservation ready", resource)

        # set correct resource status
        owner = await self.labgrid_get_place_owner(resource)
        if owner != "fc/fc":
            driver.accept_resource(resource, self)

    async def force_kick_off(self, resource):
        """
        Allow coordinator to seize labgrid resource
        """

        token = await self.labgrid_get_place_token(resource)

        if token:
            reservations = await self.labgrid_get_reservations()

            if reservations:
                for reservation in reservations.keys():
                    if reservation == f"Reservation '{token}'":
                        await self.labgrid_cancel_reservation(token)
                        await self.labgrid_release_place(resource, True)
                        break

    async def __seize_resource(self, driver, job_id, candidated_resources):
        """
        Request coordinator to seize low priority resource
        """

        if candidated_resources:
            priority_resources = await driver.coordinate_resources(
                self, job_id, *candidated_resources
            )
            if priority_resources:
                self.__update_cache("seize_cache", job_id, priority_resources)

    async def schedule(self, driver):
        """
        Monitor Labgrid reserve queue, once have pending reservation,
        release current fc acquisition to let normal user acquire the device
        Coodinator will call this function periodly
        """

        async def switch_from_fc_to_labgrid(resource):
            # if user release quickly, there possible be a window period during user release and
            # system reservation, to avoid this low probablity issue,
            # inject a low priority reservation to protect it
            await self.__labgrid_guard_reservation(resource)

            reservation = managed_resources_tokens.get(resource, None)
            if reservation:
                await self.labgrid_cancel_reservation(reservation)
            else:
                self.logger.warning(
                    "No reservation for %s found, this possible result in issue",
                    resource,
                )
            await self.labgrid_release_place(resource)

            # inject a high priority system reservation to let FC lock the device
            # after normal user finish using the device
            asyncio.create_task(self.__labgrid_system_reservation(driver, resource))

        # query labgrid reservations
        managed_resources_tokens = {}
        reservations = await self.labgrid_get_reservations()
        if reservations:
            for _, v in reservations.items():  # pylint: disable=invalid-name
                resource = v["filters"]["main"][5:]
                if v["owner"] == "fc/fc" and v["state"] in ("acquired", "allocated"):
                    managed_resources_tokens[resource] = v["token"]

                # free outdated guard reservations to prevent block due to user does quick release
                if (
                    v["owner"] == "fc/fc"
                    and v["state"] == "allocated"
                    and v["prio"] == -100.0
                ):
                    self.logger.warning(
                        "Free outdated guard reservation for %s", resource
                    )
                    await self.__labgrid_guard_reservation(resource)
                    await self.labgrid_cancel_reservation(v["token"])

            resource_list = []
            for _, v in reservations.items():  # pylint: disable=invalid-name
                resource = v["filters"]["main"][5:]
                if v["owner"] != "fc/fc" and v["state"] == "waiting":
                    if await driver.is_resource_available(self, resource):
                        if driver.is_seized_resource(self, resource):
                            driver.clear_seized_job_records(resource)

                        # if has pending reservation not belongs to normal user
                        # meanwhile device currently belongs to fc, accept it
                        driver.accept_resource(resource, self)
                        resource_list.append(resource)
                    else:
                        job_id = v["token"]

                        # pylint: disable=cell-var-from-loop
                        @check_priority_scheduler(driver)
                        @check_seize_strategy(driver, self)
                        @safe_cache
                        def labgrid_seize_resource(*_):
                            candidated_resources = (
                                []
                                if resource in self.seize_cache[job_id]
                                else [resource]
                            )

                            if (
                                not driver.is_seized_job(job_id)
                                and candidated_resources
                            ):
                                # no available resource found, try to seize from other framework
                                asyncio.create_task(
                                    self.__seize_resource(
                                        driver, job_id, candidated_resources
                                    )
                                )

                        labgrid_seize_resource(self, "seize_cache", job_id)

            if resource_list:
                await asyncio.gather(
                    *[
                        switch_from_fc_to_labgrid(resource)
                        for resource in set(resource_list)
                    ]
                )

    async def init(self, driver):
        """
        Generate and return tasks to let fc own specified labgrid devices
        This be called only once when coordinator start
        """

        places = await self.labgrid_get_places()
        self.managed_resources = [
            place for place in places if place in driver.managed_resources
        ]
        candidated_init_resources = self.managed_resources.copy()

        return [
            self.__labgrid_init(driver, resource)
            for resource in candidated_init_resources
        ]



# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: MIT


from abc import ABC, abstractmethod


class FCPlugin(ABC):
    """
    Base plugin of FC
    Detail framework plugins should realize next interfaces
    `init`, `schedule`, `force_kick_off`
    """

    def __init__(self):
        super().__init__()
        self.schedule_tick = 0
        self.schedule_interval = 1

    @abstractmethod
    async def init(self, driver):
        pass

    @abstractmethod
    async def schedule(self, driver):
        pass

    @abstractmethod
    async def force_kick_off(self, resource):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
import os
import time

import requests

from fc_common.version import get_runtime_version
from fc_server.management.cmd_online_lava_devices import LavaManagement


class Guarder:
    @staticmethod
    def ping():
        fc_server = os.environ.get("FC_SERVER", "http://127.0.0.1:8600")

        output = None
        try:
            output = requests.get(f"{fc_server}/ping")
            if output.status_code == 200 and output.text == "pong":
                return 0
        except Exception as exception:  # pylint: disable=broad-except
            print(exception)

        return 1

    @staticmethod
    def restore_lava():
        asyncio.run(LavaManagement().action())


def main():
    print("FC guarder start.")
    print(f"VERSION: {get_runtime_version('fc-guarder')}")

    default_interval = int(os.environ.get("FC_GUARDER_DEFAULT_INTERVAL", "600"))
    min_interval = int(os.environ.get("FC_GUARDER_MIN_INTERVAL", "60"))
    max_interval = int(os.environ.get("FC_GUARDER_MAX_INTERVAL", "1800"))
    max_cordon = int(os.environ.get("FC_GUARDER_MAX_CORDON", "10"))

    asleep = False
    cordon = 0
    interval = default_interval

    while True:
        ret = Guarder.ping()
        if ret == 0:
            asleep = False
            cordon = 0
            interval = default_interval
        else:
            if not asleep:
                cordon += 1
                interval = min_interval
                if cordon > max_cordon:
                    asleep = True
                    cordon = 0
                    interval = max_interval

                    # restore lava
                    print("Restore lava now.")
                    Guarder.restore_lava()
                    print("Restore lava done.")
        time.sleep(interval)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: MIT


import logging

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)



# -*- coding: utf-8 -*-
#
# Copyright 2024 NXP
#
# SPDX-License-Identifier: MIT


import os

cfg_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "fc_server", "config")
)
os.environ["FC_SERVER_CFG_PATH"] = cfg_path



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


from urllib.parse import urlparse

import etcd3
from tenacity import retry, stop_after_attempt, wait_fixed


class Etcd:
    def __init__(self, endpoint_urls, timeout=5):
        grpc_options = {
            "grpc.enable_http_proxy": 0,
        }
        endpoints = []

        for url in endpoint_urls.split(","):
            url = urlparse(url)
            endpoints.append(
                etcd3.Endpoint(
                    host=url.hostname,
                    port=url.port,
                    secure=False,
                    opts=grpc_options.items(),
                )
            )

        self.etcd = etcd3.MultiEndpointEtcd3Client(
            endpoints=endpoints, timeout=timeout, failover=True
        )

        Etcd.DeleteEvent = etcd3.events.DeleteEvent
        Etcd.PutEvent = etcd3.events.PutEvent

    def __call__(self):
        return self.etcd

    @retry(wait=wait_fixed(1), stop=stop_after_attempt(3))
    def put(self, key, value):
        return self.etcd.put(key, value)

    @retry(wait=wait_fixed(1), stop=stop_after_attempt(3))
    def get(self, key, **kwargs):
        return self.etcd.get(key, **kwargs)

    @retry(wait=wait_fixed(1), stop=stop_after_attempt(3))
    def add_watch_prefix_callback(self, key_prefix, callback, **kwargs):
        return self.etcd.add_watch_prefix_callback(key_prefix, callback, **kwargs)

    @retry(wait=wait_fixed(1), stop=stop_after_attempt(3))
    def get_prefix(self, key_prefix, **kwargs):
        return self.etcd.get_prefix(key_prefix, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
import logging

from async_lru import alru_cache

from fc_server.core.decorators import (
    check_priority_scheduler,
    check_seize_strategy,
    safe_cache,
)
from fc_server.core.plugin import FCPlugin
from fc_server.plugins.utils.lava import Lava


class Plugin(FCPlugin, Lava):
    """
    Plugin for [lava framework](https://git.lavasoftware.org/lava/lava)
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, frameworks_config):
        super().__init__()

        self.schedule_interval = 30  # poll lava job queues every 30 seconds
        accurate_scheduler_criteria = frameworks_config.get(
            "accurate_scheduler_criteria", None
        )
        self.accurate_scheduler_criteria__submitter = []
        if accurate_scheduler_criteria:
            self.accurate_scheduler_criteria__submitter = accurate_scheduler_criteria[
                "submitter"
            ]

        self.scheduler_cache = {}  # cache to avoid busy scheduling
        self.seize_cache = {}  # cache to avoid busy seize
        self.job_tags_cache = {}  # cache to store job tags

        self.logger = logging.getLogger("fc_server")

    @safe_cache
    def __update_cache(self, cache_name, job_id, value):
        self.__dict__[cache_name][job_id] += value

    async def __reset_possible_resource(self, driver, *possible_resources):
        """
        Maintenance all devices once devices finish scheduling.
        Meanwhile, return resouces which participate LAVA scheduling to FC if device idle.
        """

        # let lava scheduler schedule 90 seconds, then do corresponding cleanup
        await asyncio.sleep(90)

        if not driver.is_default_framework(self):
            await self.lava_maintenance_devices(*possible_resources)

        freed_possible_resources = []

        # check if possible resource still be used by lava
        while True:  # pylint: disable=too-many-nested-blocks
            devices = await self.lava_get_devices()
            if devices:
                used_possible_resources = [
                    device
                    for device in devices
                    if device["hostname"] in possible_resources
                    and device["hostname"] not in freed_possible_resources
                ]

                if len(used_possible_resources) == 0:
                    break

                for info in used_possible_resources:
                    if not info["current_job"]:
                        await driver.return_resource(info["hostname"])
                        freed_possible_resources.append(info["hostname"])

                        # clean cache for returned device
                        for job_id in list(self.scheduler_cache.keys()):
                            if info["hostname"] in self.scheduler_cache[job_id]:
                                del self.scheduler_cache[job_id]
                        for job_id in list(self.seize_cache.keys()):
                            if info["hostname"] in self.seize_cache[job_id]:
                                del self.seize_cache[job_id]
            await asyncio.sleep(60)

    async def __get_job_tags(self, job_id):
        """
        Return job tag info, issue interface call will return None
        """

        job_info = await self.lava_get_job_info(job_id)
        return (job_id, job_info["tags"]) if job_info else None

    async def __get_device_tags(self, device):
        """
        Return device tag info, issue interface call will return None
        """

        device_info = await self.lava_get_device_info(device)
        return (device, device_info["tags"]) if device_info else None

    async def __get_device_info(self, device, clear=False):
        """
        Configureable wrapper of lava_get_device_info to clean cache
        """

        if clear:
            self.__get_cached_device_info.cache_clear()  # pylint: disable=no-member

        return await self.__get_cached_device_info(device)

    @alru_cache(None)
    async def __get_cached_device_info(self, device):
        """
        Cached wrapper of lava_get_device_info
        """

        return await self.lava_get_device_info(device)

    async def __get_device_description(self, device):
        """
        Return device description, issue interface call will return None
        """

        device_info = await self.__get_device_info(device)
        return (
            device_info["description"] if device_info else self.lava_default_description
        )

    async def force_kick_off(self, resource):
        """
        Allow coordinator to seize lava resource
        """

        device_info = await self.lava_get_device_info(resource)

        if not device_info:
            return

        current_job = device_info["current_job"]

        if current_job:
            await self.lava_cancel_job(current_job)

    async def __seize_resource(self, driver, job_id, candidated_non_available_devices):
        """
        Request coordinator to seize low priority resource
        """

        non_available_device_tags_list = await asyncio.gather(
            *[
                self.__get_device_tags(candidated_non_available_device)
                for candidated_non_available_device in candidated_non_available_devices
            ]
        )
        non_available_device_tags_list = filter(
            lambda _: isinstance(_, tuple), non_available_device_tags_list
        )
        non_available_device_tags_dict = dict(non_available_device_tags_list)

        candidated_non_available_resources = []
        for device, tags in non_available_device_tags_dict.items():
            if set(self.job_tags_cache[job_id]).issubset(tags):
                candidated_non_available_resources.append(device)
            else:
                self.__update_cache("seize_cache", job_id, [device])

        if candidated_non_available_resources:
            priority_resources = await driver.coordinate_resources(
                self, job_id, *candidated_non_available_resources
            )
            if priority_resources:
                self.__update_cache("seize_cache", job_id, priority_resources)

    async def default_framework_disconnect(self, resource):
        """
        Default framework should realize this to let FC control the disconnect
        """

        device_info = await self.__get_device_info(resource, clear=True)
        if not device_info or device_info["current_job"]:
            return False, False

        if device_info["health"] == "Maintenance":
            self.logger.info("%s default in maintenance.", resource)
            return True, False

        if device_info["health"] == "Retired":
            self.logger.info("%s default in retired.", resource)
            return False, False

        # ask default framework disconnect this resource
        self.logger.info("Disconnect %s from default framework", resource)
        desc = await self.__get_device_description(resource)
        if not await self.lava_maintenance_devices(
            resource, desc=f"{self.device_description_prefix}{desc}"
        ):
            return False, False

        device_info = await self.__get_device_info(resource, clear=True)
        if not device_info or device_info["current_job"]:
            return False, True

        return True, True

    async def default_framework_connect(self, resource):
        """
        Default framework should realize this to let FC control the connect
        """

        # ask default framework connect this resource
        self.logger.info("Connect %s to default framework", resource)
        desc = await self.__get_device_description(resource)
        return await self.lava_online_devices(
            resource, desc=desc.split(self.device_description_prefix)[-1]
        )

    async def schedule(
        self, driver
    ):  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        """
        Monitor LAVA job queue, once have pending jobs, online related devices
        to let LAVA take over these devices
        Coodinator will call this function periodly
        """

        async def schedule_prepare():
            # retire managed resources
            devices_assemble = [device["hostname"] for device in devices]
            retired_managed_resources = list(
                set(driver.managed_resources).difference(set(devices_assemble))
            )

            for retired_managed_resource in retired_managed_resources:
                if (
                    driver.managed_resources_status[retired_managed_resource]
                    != "retired"
                ):
                    driver.retire_resource(retired_managed_resource)

            for managed_resource in driver.managed_resources:
                if (
                    driver.managed_resources_status[managed_resource] == "retired"
                    and managed_resource in devices_assemble
                ):
                    driver.reset_resource(managed_resource)

            # category managed resources
            for device in devices:
                if device["hostname"] in driver.managed_resources and device[
                    "health"
                ] in (
                    ("Maintenance", "Unknown", "Good", "Bad")
                    if not driver.is_default_framework(self)
                    else (
                        ("Unknown", "Good", "Maintenance")
                        if driver.managed_disconnect_resource(device["hostname"])
                        else ("Unknown", "Good")
                    )
                ):
                    if not driver.is_default_framework(self):
                        # yield resource which should assure in maintenance mode
                        if device["health"] in (
                            "Unknown",
                            "Good",
                            "Bad",
                        ) and await driver.is_resource_available(
                            self, device["hostname"]
                        ):
                            yield device["hostname"]

                    # guard behavior: in case there are some unexpected manual online & jobs there
                    if device["current_job"] and await driver.is_resource_available(
                        self, device["hostname"]
                    ):
                        driver.accept_resource(device["hostname"], self)
                        asyncio.create_task(
                            self.__reset_possible_resource(
                                driver, *(device["hostname"],)
                            )
                        )

                    # category devices by devicetypes as LAVA schedule based on devicetypes
                    if await driver.is_resource_available(self, device["hostname"]):
                        category_key = "available"
                    elif driver.is_resource_non_available(device["hostname"]):
                        category_key = "non-available"

                    if device["type"] in managed_resources_category[category_key]:
                        managed_resources_category[category_key][device["type"]].append(
                            device["hostname"]
                        )
                    else:
                        managed_resources_category[category_key][device["type"]] = [
                            device["hostname"]
                        ]

        # category devices
        managed_resources_category = {"available": {}, "non-available": {}}
        devices = await self.lava_get_devices()

        if not devices:
            self.logger.warning(
                "No device fetched from lava server, delay to next scheduling slot"
            )
            return

        if driver.is_default_framework(self):
            async for _ in schedule_prepare():
                pass
        else:
            await self.lava_maintenance_devices(schedule_prepare())

        # query job queue
        possible_resources = []
        queued_jobs = await self.lava_get_queued_jobs()

        # clean cache to save memory
        queued_jobs_ids = [queued_job["id"] for queued_job in queued_jobs]
        for job_id in list(self.job_tags_cache.keys()):
            if job_id not in queued_jobs_ids:
                del self.job_tags_cache[job_id]

        # get tags for queued jobs
        job_tags_list = await asyncio.gather(
            *[
                self.__get_job_tags(queued_job["id"])
                for queued_job in queued_jobs
                if queued_job["id"] not in self.job_tags_cache
            ]
        )
        job_tags_list = filter(lambda _: isinstance(_, tuple), job_tags_list)
        self.job_tags_cache.update(dict(job_tags_list))

        # get devices suitable for queued jobs
        queued_jobs.reverse()
        for queued_job in queued_jobs:
            job_id = queued_job["id"]

            # delay issue job to next scheduling slot
            if job_id not in self.job_tags_cache:
                self.logger.warning("Job %s delayed to next scheduling slot", job_id)
                continue

            candidated_available_devices = managed_resources_category["available"].get(
                queued_job["requested_device_type"], []
            )

            candidated_available_resources = []

            if job_id not in self.scheduler_cache:
                self.scheduler_cache[job_id] = []

            available_device_tags_list = await asyncio.gather(
                *[
                    self.__get_device_tags(candidated_available_device)
                    for candidated_available_device in candidated_available_devices
                    if candidated_available_device not in self.scheduler_cache[job_id]
                ]
            )
            available_device_tags_list = filter(
                lambda _: isinstance(_, tuple), available_device_tags_list
            )
            available_device_tags_dict = dict(available_device_tags_list)

            for device, tags in available_device_tags_dict.items():
                if set(self.job_tags_cache[job_id]).issubset(tags):
                    candidated_available_resources.append(device)

                    if driver.is_seized_resource(self, device):
                        driver.clear_seized_job_records(device)

            self.__update_cache(
                "scheduler_cache", job_id, list(available_device_tags_dict.keys())
            )
            possible_resources += candidated_available_resources

            # pylint: disable=cell-var-from-loop
            @check_priority_scheduler(driver)
            @check_seize_strategy(driver, self)
            @safe_cache
            def lava_seize_resource(*_):
                candidated_non_available_devices = [
                    non_available_device
                    for non_available_device in managed_resources_category[
                        "non-available"
                    ].get(queued_job["requested_device_type"], [])
                    if non_available_device not in self.seize_cache[job_id]
                ]

                if (
                    not candidated_available_resources
                    and not driver.is_seized_job(job_id)
                    and candidated_non_available_devices
                ):
                    # no available resource found, try to seize from other framework
                    asyncio.create_task(
                        self.__seize_resource(
                            driver, job_id, candidated_non_available_devices
                        )
                    )

            lava_seize_resource(self, "seize_cache", job_id)

        possible_resources = set(possible_resources)

        # let lava dispatch
        if possible_resources:
            for possible_resource in possible_resources:
                driver.accept_resource(possible_resource, self)

            if not driver.is_default_framework(self):
                self.logger.info("Online devices to schedule lava jobs.")
                await self.lava_online_devices(*possible_resources)

            # cleanup
            asyncio.create_task(
                self.__reset_possible_resource(driver, *possible_resources)
            )

    async def init(self, driver):
        """
        Generate and return tasks to let fc own specified lava devices correctly
        Called only once when coordinator start
        """

        if driver.is_default_framework(self):
            return []

        return [
            self.lava_maintenance_devices(device["hostname"])
            for device in await self.lava_get_devices()
            if device["hostname"] in driver.managed_resources
            and device["health"] in ("Unknown", "Good", "Bad")
        ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2021-2024 NXP
#
# SPDX-License-Identifier: MIT


import argparse
import asyncio
import json
import os
import pathlib
import signal
import socket
import subprocess
import sys
import time
import urllib.parse
from getpass import getuser
from socket import gethostname

import aiohttp
import prettytable
import psutil
import requests
import yaml

from fc_common import which
from fc_common.config import Config
from fc_common.version import get_runtime_version


class Client:
    daemon_pid_file = "/tmp/fc/fc_client_daemon.pid"
    etcd_url = None

    @staticmethod
    def mode_check():
        fc_server = os.environ.get("FC_SERVER", None)
        lg_crossbar = os.environ.get("LG_CROSSBAR", None)

        if fc_server and lg_crossbar:
            print("MODE: single")
            Client.mode = "single"
        elif fc_server:
            print("MODE: cluster (LG_CROSSBAR not set, fallback to cluster mode)")
            Client.mode = "cluster"
        elif lg_crossbar:
            print("MODE: cluster (FC_SERVER not set, fallback to cluster mode)")
            Client.mode = "cluster"
        else:
            print("MODE: cluster")
            Client.mode = "cluster"

    @staticmethod
    def labgrid_call(args, extras):
        metadata = Client.fetch_metadata(args.resource)
        os.environ["LG_CROSSBAR"] = metadata["lg"]

        # FIXME: temp fix for bootstrap command  # pylint: disable=fixme
        if extras == ["ssh"]:
            cmd = " ".join(extras)
            try:
                import fc_plugins  # pylint: disable=import-outside-toplevel

                cmd = fc_plugins.get_rule(metadata["lg"], cmd)
            except ImportError:
                pass

            os.environ["LG_PLACE"] = args.resource
            os.execvp(
                "labgrid-client",
                ["labgrid-client", "-p", args.resource] + cmd.split(" "),
            )
        else:
            os.execvp(
                "labgrid-client", ["labgrid-client", "-p", args.resource] + extras
            )

    @staticmethod
    def communicate_with_daemon(msg_type, para=None):
        if os.path.exists(Client.daemon_pid_file):
            pid_file = pathlib.Path(Client.daemon_pid_file)
            pid_file = pid_file.resolve()

            pid = int(pid_file.read_text(encoding="utf-8").rstrip())

            if not psutil.pid_exists(pid):
                os.remove(Client.daemon_pid_file)

        if not os.path.exists(Client.daemon_pid_file):
            client_daemon = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "..",
                "fc_client_daemon",
                "client_daemon.py",
            )
            subprocess.run([sys.executable, client_daemon], check=True)

        server_address = "\0/tmp/fc/fc_client_daemon.sock"
        retries = 0
        max_retries = 50
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        while True:
            try:
                sock.connect(server_address)
                break
            except socket.error as msg:
                if retries == max_retries:
                    sys.exit(f"Fatal: fc_client_daemon not available ({msg})")
                time.sleep(0.1)
                retries += 1

        msg = {"msg_type": msg_type, "para": para}
        json_msg = json.dumps(msg)

        sock.send(json_msg.encode("utf-8"))
        data = sock.recv(1024)
        sock.close()

        return data

    @staticmethod
    def fetch_metadata(filters):
        # single mode
        if Client.mode == "single":
            fc_server = os.environ.get("FC_SERVER", None)
            lg_crossbar = os.environ.get("LG_CROSSBAR", None)

            if filters == "all":
                return {"default": {"fc": fc_server, "lg": lg_crossbar}}
            return {"fc": fc_server, "lg": lg_crossbar}

        # cluster mode
        def check_etcd_cfg():
            Client.etcd_url = Config.load_cfg().get("etcd")
            if not Client.etcd_url:
                sys.exit("Fatal: please init cluster settings for your client first")

        check_etcd_cfg()

        data = Client.communicate_with_daemon("require_info", filters)
        json_data = json.loads(data.decode("utf-8"))

        if filters != "all" and not json_data:
            sys.exit("Fatal: invalid resource")

        return json_data

    @staticmethod
    def init(extras):
        if len(extras) == 0 or extras[0] not in ["etcd"]:
            sys.exit("Candidated init para: etcd")

        if len(extras) == 1:
            cfg = Config.load_cfg()
            cfg.pop(extras[0], "")
            Config.save_cfg(cfg)
            Client.communicate_with_daemon("daemon_stop")
            print(f"{extras[0]} removed")
        elif len(extras) == 2:
            cfg = Config.load_cfg()
            cfg.update({extras[0]: extras[1]})
            Config.save_cfg(cfg)
            Client.communicate_with_daemon("daemon_stop")
            print("Init done")
        else:
            print("Wrong command")

    @staticmethod
    def cluster_info(args):
        if args.resource:
            metadata = Client.fetch_metadata(args.resource)
            print(f"export FC_SERVER={metadata['fc']}")
            print(f"export LG_CROSSBAR={metadata['lg']}")
        else:
            metadata = Client.fetch_metadata("all")
            print(f"ETCD: {Client.etcd_url}")

            print(f"Totally {len(metadata)} instances as follows:")
            for instance_name, data in metadata.items():
                print(f"{instance_name}:")
                print(f"  - export FC_SERVER={data['fc']}")
                print(f"  - export LG_CROSSBAR={data['lg']}")

    @staticmethod
    def booking(_):
        async def get_booking(fc_server):
            url = f"{fc_server}/booking"
            timeout = aiohttp.ClientTimeout(total=5)
            async with aiohttp.ClientSession(timeout=timeout) as session:
                try:
                    async with session.get(url) as response:
                        output = await response.text()
                except Exception as exc:  # pylint: disable=broad-except
                    print(f"Warning: {fc_server} skipped due to {type(exc)}")
                    return ""

            return output

        metadata = Client.fetch_metadata("all")

        instances = list(metadata.keys())
        tasks = [
            get_booking(instance_meta["fc"]) for instance_meta in metadata.values()
        ]
        loop = asyncio.get_event_loop()
        booking_data = loop.run_until_complete(asyncio.gather(*tasks))

        for index, data in enumerate(booking_data):
            print(f"- Instance {instances[index]}:")
            print(data)

    @staticmethod
    def status(args):  # pylint: disable=too-many-statements
        async def get_status(fc_server):
            specified_resource = args.resource
            specified_farm_type = args.farm_type
            specified_device_type = args.device_type
            specified_peripheral_info = args.peripheral_info

            if args.verbose == 0:
                url = f"{fc_server}/resource"
            else:
                url = f"{fc_server}/verbose_resource"

            if specified_resource:
                url += f"/{specified_resource}"

            query = {}
            if specified_farm_type:
                query["farmtype"] = specified_farm_type
            if specified_device_type:
                query["devicetype"] = specified_device_type
            if specified_peripheral_info:
                query["peripheralinfo"] = specified_peripheral_info
            query_string = urllib.parse.urlencode(query)

            if query_string:
                url += f"?{query_string}"

            timeout = aiohttp.ClientTimeout(total=5)
            async with aiohttp.ClientSession(timeout=timeout) as session:
                try:
                    async with session.get(url) as response:
                        output = await response.text()
                except Exception as exc:  # pylint: disable=broad-except
                    print(f"Warning: {fc_server} skipped due to {type(exc)}")
                    return []

            try:
                ret = json.loads(output)
            except Exception:  # pylint: disable=broad-except
                return []

            return ret

        metadata = Client.fetch_metadata("all")

        tasks = [get_status(instance_meta["fc"]) for instance_meta in metadata.values()]
        loop = asyncio.get_event_loop()
        resource_data = loop.run_until_complete(asyncio.gather(*tasks))
        resources = sum(resource_data, [])

        table = prettytable.PrettyTable(align="l")

        if args.verbose == 0:
            table_width = 4
            for resource in resources:
                table.add_row(resource)
                table_width = len(resource)

            if table_width == 5:
                table.field_names = ["Resource", "Farm", "Status", "Note", "Info"]
            else:
                table.field_names = ["Resource", "Farm", "Status", "Note"]
        else:
            table_width = 5
            for resource in resources:
                table.add_row(resource)
                table_width = len(resource)

            if table_width == 6:
                table.field_names = [
                    "Resource",
                    "Farm",
                    "Status",
                    "Note",
                    "Comment",
                    "Info",
                ]
            else:
                table.field_names = ["Resource", "Farm", "Status", "Note", "Comment"]

        table.max_width = 80
        print(table.get_string(sortby="Resource"))

    @staticmethod
    @which(
        "labgrid-client",
        "Use 'pip3 install labgrid' or other customized way to install labgrid software please.",
    )
    def lock(args):
        resource = args.resource
        metadata = Client.fetch_metadata(resource)

        fc_server = metadata.get("fc", None)
        if not fc_server:
            sys.exit("Fatal: invalid resource")
        else:
            url = f"{fc_server}/resource/{resource}"
            output = requests.get(url)
            output_data = json.loads(output.text)

            if len(output_data) == 0:
                sys.exit("Fatal: invalid resource")

            if output_data[0][3] != "":
                sys.exit("Fatal: non-debuggable resource")

        os.environ["LG_CROSSBAR"] = metadata["lg"]

        if resource:
            print(f"Check if {resource} is free")
            cmd = f"labgrid-client -p {resource} show | grep acquired:"
            ret, place_info_text = subprocess.getstatusoutput(cmd)
            if ret == 0:
                acquired = place_info_text.split()[1]
                if acquired not in ["fc/fc", "None"]:
                    sys.exit(
                        f"error: place {resource} is already acquired by {acquired}"
                    )
            print(f"Try to acquire resource {resource}...")
            with subprocess.Popen(
                ["labgrid-client", "reserve", "--wait", f"name={resource}"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as process:
                try:
                    process.communicate()
                    subprocess.call(["labgrid-client", "-p", resource, "lock"])
                except KeyboardInterrupt:
                    signal.signal(signal.SIGINT, lambda _: "")
                    token = ""
                    for line in process.stdout.readlines():
                        line = line.decode("UTF-8").strip()
                        if line.startswith("token:"):
                            token = line[7:]
                            break
                    if token:
                        subprocess.call(["labgrid-client", "cancel-reservation", token])
                        subprocess.call(
                            ["labgrid-client", "-p", resource, "unlock"],
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.STDOUT,
                        )
        else:
            sys.exit("No resource specified.")

    @staticmethod
    @which(
        "labgrid-client",
        "Use 'pip3 install labgrid' or other customized way to install labgrid software please.",
    )
    def unlock(args):
        resource = args.resource
        metadata = Client.fetch_metadata(resource)
        os.environ["LG_CROSSBAR"] = metadata["lg"]

        me = "/".join(  # pylint: disable=invalid-name
            (
                os.environ.get("LG_HOSTNAME", gethostname()),
                os.environ.get("LG_USERNAME", getuser()),
            )
        )

        if resource:  # pylint: disable=too-many-nested-blocks
            cmd = f"labgrid-client -p {resource} show"
            ret, place_info_text = subprocess.getstatusoutput(cmd)
            if ret == 0:
                token = ""
                place_info_lines = place_info_text.splitlines()
                for line in place_info_lines:
                    if line.find("reservation") >= 0:
                        token = line.split(":")[-1].strip()
                        break

                if token:
                    cmd = "labgrid-client reservations"
                    reservations_text = subprocess.check_output(cmd, shell=True)
                    reservations = yaml.load(reservations_text, Loader=yaml.FullLoader)
                    for k, v in reservations.items():  # pylint: disable=invalid-name
                        if k == f"Reservation '{token}'":
                            owner = v["owner"]
                            if owner == me:
                                print("Start to free the place.")
                                signal.signal(signal.SIGINT, lambda _: "")
                                subprocess.call(
                                    ["labgrid-client", "cancel-reservation", token]
                                )
                                subprocess.call(
                                    ["labgrid-client", "-p", resource, "unlock"]
                                )
                            else:
                                sys.exit("Fatal: the resource not owned by you.")
                            break
            else:
                sys.exit(f"Fatal: {place_info_text}")
        else:
            sys.exit("No resource specified.")


def main():
    print(
        f"FC-CLIENT VERSION: {get_runtime_version('fc-client')}, "
        "HOMEPAGE: https://fc.readthedocs.org/"
    )

    Client.mode_check()

    parser = argparse.ArgumentParser()
    parser.set_defaults(func=lambda args: parser.print_help())
    parser.add_argument(
        "-r", "--resource", "-p", "--place", type=str, help="resource name"
    )
    parser.add_argument("-f", "--farm-type", type=str, help="farm type")
    parser.add_argument("-d", "--device-type", type=str, help="device type")
    parser.add_argument("-i", "--peripheral-info", type=str, help="peripheral info")
    parser.add_argument("-v", "--verbose", action="count", default=0)

    args, extras = parser.parse_known_args()

    if (
        len(extras) > 0
        and extras[0]
        in [
            "status",
            "s",
            "lock",
            "l",
            "unlock",
            "u",
            "booking",
            "b",
            "cluster-info",
            "c",
            "init",
            "i",
        ]
        or not args.resource
    ):
        subparsers = parser.add_subparsers(
            dest="command",
            title="available subcommands",
            metavar="COMMAND",
        )

        subparser = subparsers.add_parser(
            "status", aliases=("s",), help="list status of fc resource"
        )
        subparser.set_defaults(func=Client.status)

        subparser = subparsers.add_parser(
            "lock", aliases=("l", "acquire"), help="labgrid lock resource"
        )
        subparser.set_defaults(func=Client.lock)

        subparser = subparsers.add_parser(
            "unlock", aliases=("u", "release"), help="labgrid unlock resource"
        )
        subparser.set_defaults(func=Client.unlock)

        subparser = subparsers.add_parser(
            "booking", aliases=("b",), help="list current booking"
        )
        subparser.set_defaults(func=Client.booking)

        subparser = subparsers.add_parser(
            "cluster-info", aliases=("c",), help="list cluster info"
        )
        subparser.set_defaults(func=Client.cluster_info)

        subparser = subparsers.add_parser("init", aliases=("i",), help="init fc-client")
        subparser.set_defaults(func=Client.init)

        if len(extras) > 0 and extras[0] in ["init", "i"]:
            args, extras = parser.parse_known_args(extras, namespace=args)
            args.func(extras)
        else:
            args = parser.parse_args(extras, namespace=args)
            args.func(args)
    else:
        Client.labgrid_call(args, extras)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


import os

import yaml


class Config:
    @staticmethod
    def save_cfg(cfg):
        cfg_path = os.path.expanduser(
            os.environ.get("FC_CLIENT_CFG_PATH", "~/.fc/client_cfg")
        )
        cfg_yaml = os.path.join(cfg_path, "fc.yaml")
        os.makedirs(cfg_path, exist_ok=True)

        with open(cfg_yaml, "w", encoding="utf-8") as f_cfg:
            yaml.dump(cfg, f_cfg)

    @staticmethod
    def load_cfg():
        system_cfg_path = "/opt/.fc/client_cfg"
        system_cfg_yaml = os.path.join(system_cfg_path, "fc.yaml")
        if os.path.exists(system_cfg_yaml):
            cfg_yaml = system_cfg_yaml
        else:
            cfg_path = os.path.expanduser(
                os.environ.get("FC_CLIENT_CFG_PATH", "~/.fc/client_cfg")
            )
            cfg_yaml = os.path.join(cfg_path, "fc.yaml")

        try:
            with open(cfg_yaml, encoding="utf-8") as f_cfg:
                data = yaml.load(f_cfg.read(), yaml.SafeLoader)
                return data
        except Exception:  # pylint: disable=broad-except
            return {}



# Configuration file for the Sphinx documentation builder.

# -- Project information

import subprocess

project = 'FC'
copyright = '2021-2023, NXP, Larry Shen'
author = 'Larry Shen'

version = (
    subprocess.Popen(
        ["cat", "fc_common/VERSION"], cwd=r"..", stdout=subprocess.PIPE
    )
    .stdout.read()
    .rstrip()
    .decode("utf-8")
)
release = version

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.imgconverter',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for EPUB output

epub_show_urls = 'footnote'

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

html_logo = 'images/fc.png'

html_theme_options = {
    'collapse_navigation': False,
}

html_static_path = ['css']

html_css_files = [
    'css-style.css',
]



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
import os
import sys

import pytest

cfg_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "config"))
os.environ["FC_SERVER_CFG_PATH"] = cfg_path

# pylint: disable=wrong-import-position
from fc_server.core.coordinator import Coordinator


@pytest.fixture
def remove_config_path():
    del os.environ["FC_SERVER_CFG_PATH"]
    yield
    os.environ["FC_SERVER_CFG_PATH"] = cfg_path


@pytest.fixture
def coordinator():
    return Coordinator()


@pytest.fixture
def asyncio_patch():
    def _asyncio_patch(mocker, target, result):
        if sys.version_info >= (3, 8):
            from unittest.mock import (  # pylint: disable=import-outside-toplevel
                AsyncMock,
            )

            async_mock = AsyncMock(return_value=result)
            return mocker.patch(target, side_effect=async_mock)

        future = asyncio.Future()
        future.set_result(result)
        return mocker.patch(target, return_value=future)

    return _asyncio_patch



# -*- coding: utf-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
import logging
import sys
from importlib import import_module

from fc_server.core.api_svr import ApiSvr
from fc_server.core.config import Config
from fc_server.core.decorators import check_priority_scheduler
from fc_server_daemon.server_daemon import ServerDaemon


class Coordinator:
    """
    FC coordinator which used to coordinate status among different frameworks
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self):
        self.logger = logging.getLogger("fc_server")

        self.__framework_plugins = [
            import_module("fc_server.plugins." + framework).Plugin(
                Config.frameworks_config[framework]
            )
            for framework in Config.registered_frameworks
        ]

        self.__framework_priorities = {
            framework: Config.frameworks_config[framework]["priority"]
            for framework in Config.registered_frameworks
        }

        self.__framework_seize_strategies = {
            framework: Config.frameworks_config[framework].get("seize", True)
            for framework in Config.registered_frameworks
        }

        if Config.default_framework:
            self.logger.info("Default framework: %s", Config.default_framework)
            for framework in self.__framework_plugins:
                if framework.__module__.split(".")[-1] == Config.default_framework:
                    self.__default_framework_plugin = framework
                    if not hasattr(
                        self.__default_framework_plugin, "default_framework_disconnect"
                    ) or not hasattr(
                        self.__default_framework_plugin, "default_framework_connect"
                    ):
                        self.logger.fatal(
                            "Fatal: specified default framework doesn't realize next interfaces:"
                        )
                        self.logger.fatal("  - default_framework_disconnect")
                        self.logger.fatal("  - default_framework_connect")
                        sys.exit(1)
                    break

        self.__managed_resources_status = {}
        self.__managed_disconnect_resources = []
        self.__managed_issue_disconnect_resources = []

        for resource in Config.managed_resources:
            self.__managed_resources_status[resource] = "fc"

        # assure no seize for this job when already a seize for this job there
        self.coordinating_job_records = {}

        # record timeout task for seized status
        self.seized_status_timeout_records = {}

        self.logger.info("FC managed resource list:")
        self.logger.info(Config.managed_resources)

    async def __init_frameworks(self):
        self.logger.info("Start to init following frameworks:")

        init_tasks = []
        for framework in self.__framework_plugins:
            self.logger.info("  - %s", framework.__module__)
            init_tasks += await framework.init(self)

        await asyncio.gather(*init_tasks)

        self.logger.info("Framework coordinator ready.")

    async def __managed_issue_resources_connect(self):
        for resource in self.__managed_issue_disconnect_resources:
            if await self.__default_framework_plugin.default_framework_connect(
                resource
            ):
                self.__managed_issue_disconnect_resources.remove(resource)

    async def __schedule_frameworks(self):
        while True:
            # connect all resources which disconnect by fc but previously not successful connect
            await self.__managed_issue_resources_connect()

            # schedule different frameworks
            for framework in self.__framework_plugins:
                if framework.schedule_tick % framework.schedule_interval == 0:
                    framework.schedule_tick = 0
                    await framework.schedule(self)
                framework.schedule_tick += 1

            await asyncio.sleep(1)

    async def __action(self):
        await self.__init_frameworks()
        await ApiSvr(self).start(**Config.api_server)
        await self.__schedule_frameworks()

    @property
    def priority_scheduler(self):
        return Config.priority_scheduler

    @property
    def managed_resources(self):
        return Config.managed_resources

    @property
    def managed_resources_status(self):
        return self.__managed_resources_status

    @property
    def framework_instances(self):
        return self.__framework_plugins

    @property
    def framework_seize_strategies(self):
        return self.__framework_seize_strategies

    def managed_disconnect_resource(self, resource):
        return resource in self.__managed_disconnect_resources

    def is_default_framework(self, context):  # pylint: disable=no-self-use
        return context.__module__.split(".")[-1] == Config.default_framework

    def __get_low_priority_frameworks(self, cur_framework):
        """
        Get all frameworks with low priority compared to current framework
        """

        return [
            framework
            for framework in self.__framework_priorities.keys()
            if self.__framework_priorities[framework]
            > self.__framework_priorities[cur_framework]
        ]

    async def is_resource_available(self, context, resource):
        if self.__managed_resources_status.get(resource, "") == "fc":
            if Config.default_framework:
                if context.__module__.split(".")[-1] == Config.default_framework:
                    return True

                (
                    disconnect_success,
                    maintenance_by_fc,
                ) = await self.__default_framework_plugin.default_framework_disconnect(
                    resource
                )

                if maintenance_by_fc:
                    if not disconnect_success:
                        # delay default framework connect if race condition
                        self.__managed_issue_disconnect_resources.append(resource)
                return disconnect_success

            return True

        if (
            self.__managed_resources_status.get(resource, "")
            == context.__module__.split(".")[-1] + "_seized"
        ):
            return True

        return False

    def is_resource_non_available(self, resource):
        return (
            self.__managed_resources_status.get(resource, "")
            in Config.registered_frameworks
        )

    @check_priority_scheduler()
    def is_seized_resource(self, context, resource):
        return (
            self.__managed_resources_status.get(resource, "")
            == context.__module__.split(".")[-1] + "_seized"
        )

    def clear_seized_job_records(self, device):
        for job_id in list(self.coordinating_job_records.keys()):
            if device == self.coordinating_job_records[job_id]:
                self.coordinating_job_records.pop(job_id)

    def is_seized_job(self, job_id):
        # pylint: disable=consider-iterating-dictionary
        return job_id in list(self.coordinating_job_records.keys())

    async def __seized_status_timeout(self, resource):
        await asyncio.sleep(90)
        self.logger.info("* %s seized status be reset to fc due to timeout", resource)
        self.reset_resource(resource)

    @check_priority_scheduler()
    async def coordinate_resources(self, context, job_id, *candidated_resources):
        """
        Seize resource from low priority framework
        """

        if not candidated_resources:
            return []

        self.logger.info(
            "[start] seize resource requirement from %s for %s",
            context.__module__.split(".")[-1],
            job_id,
        )

        candidated_seized_resources = [
            candidated_resource
            for candidated_resource in candidated_resources
            if self.managed_resources_status[candidated_resource]
            in self.__get_low_priority_frameworks(context.__module__.split(".")[-1])
        ]

        high_priority_resources = list(
            set(candidated_resources).difference(set(candidated_seized_resources))
        )
        low_priority_resources = []

        if candidated_seized_resources:
            candidated_seized_resource = candidated_seized_resources[0]
            self.coordinating_job_records[job_id] = candidated_seized_resource

            # kick off resource
            for framework in self.__framework_plugins:
                if (
                    framework.__module__.split(".")[-1]
                    == self.managed_resources_status[candidated_seized_resource]
                ):
                    self.__set_resource_status(
                        candidated_seized_resource,
                        context.__module__.split(".")[-1] + "_seizing",
                    )
                    self.logger.info(
                        "Force kick off the resource %s.", candidated_seized_resource
                    )
                    await framework.force_kick_off(candidated_seized_resource)
                    self.__set_resource_status(
                        candidated_seized_resource,
                        context.__module__.split(".")[-1] + "_seized",
                    )

                    # timeout for seized status
                    timeout_task = asyncio.create_task(
                        self.__seized_status_timeout(candidated_seized_resource)
                    )
                    self.seized_status_timeout_records[
                        candidated_seized_resource
                    ] = timeout_task

                    break
            low_priority_resources.append(candidated_seized_resource)

        self.logger.info(
            "[done] seize resource requirement from %s for %s",
            context.__module__.split(".")[-1],
            job_id,
        )

        return high_priority_resources + low_priority_resources

    def __set_resource_status(self, resource, status):
        self.__managed_resources_status[resource] = status
        self.logger.info("* %s now belongs to %s", resource, status)

    async def return_resource(self, resource):
        if (
            self.__managed_resources_status.get(resource, "")
            in Config.registered_frameworks
        ):
            self.__set_resource_status(resource, "fc")

        if Config.default_framework and resource in self.__managed_disconnect_resources:
            # restore default framework resource status
            self.__managed_disconnect_resources.remove(resource)
            if not await self.__default_framework_plugin.default_framework_connect(
                resource
            ):
                # delay default framework connect for this resource if connect api call failure
                self.__managed_issue_disconnect_resources.append(resource)

    def accept_resource(self, resource, context):
        # cancel seized status timeout task if this resource is seized from others
        if resource in self.seized_status_timeout_records:
            timeout_task = self.seized_status_timeout_records.pop(resource)
            timeout_task.cancel()

        self.__set_resource_status(resource, context.__module__.split(".")[-1])
        if Config.default_framework:
            if context.__module__.split(".")[-1] != Config.default_framework:
                self.__managed_disconnect_resources.append(resource)

    def retire_resource(self, resource):
        self.__set_resource_status(resource, "retired")

    def reset_resource(self, resource):
        self.__set_resource_status(resource, "fc")

    def start(self):
        if Config.cluster["enable"]:
            daemon_paras = {
                "etcd": Config.cluster["etcd"],
                "instance_name": Config.cluster["instance_name"],
                "fc": f"http://{Config.api_server['ip']}:{Config.api_server['publish_port']}",
                "lg": Config.frameworks_config["labgrid"]["lg_crossbar"],
                "devices": Config.managed_resources,
            }
            ServerDaemon(daemon_paras).run()

        asyncio.run(self.__action())



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


from unittest.mock import MagicMock

import pytest

from fc_server.plugins.lava import Plugin


@pytest.fixture(name="plugin")
def lava_plugin():
    config = {"identities": "$lava_identity", "priority": 1, "default": True}
    return Plugin(config)


@pytest.fixture(name="lava_job_info")
def create_lava_job_info():
    job_info = {
        "description": "foo",
        "device": None,
        "device_type": "docker",
        "end_time": None,
        "Health": "Unknown",
        "health_check": False,
        "id": "0",
        "pipeline": True,
        "start_time": None,
        "state": "Submitted",
        "submit_time": "20181101T07:13:06",
        "submitter": "foobar",
        "tags": [],
        "visibility": "Public",
    }
    return job_info


@pytest.fixture(name="lava_device_info")
def create_lava_device_info():
    device_info = {
        "current_job": None,
        "description": "Created automatically by LAVA.",
        "device_type": "docker",
        "has_device_dict": "true",
        "health": "Unknown",
        "health_job": "false",
        "hostname": "$resource1",
        "pipeline": "true",
        "state": "Idle",
        "tags": [],
        "worker": "foobar",
    }
    return device_info


@pytest.fixture(name="lava_device_info_unknown")
def create_lava_device_info_unknown(lava_device_info):
    return lava_device_info


@pytest.fixture(name="lava_device_info_good")
def create_lava_device_info_good(lava_device_info):
    lava_device_info.update({"health": "Good"})
    return lava_device_info


@pytest.fixture(name="lava_device_info_with_job")
def create_lava_device_info_with_job(lava_device_info):
    lava_device_info.update({"current_job": "1"})
    return lava_device_info


# pylint: disable=protected-access
class TestPluginLava:
    def test_update_cache(self, plugin):
        plugin._Plugin__update_cache("scheduler_cache", "0", ["foo"])
        assert plugin.scheduler_cache["0"] == ["foo"]

    @pytest.mark.asyncio
    async def test_get_job_tags(self, asyncio_patch, mocker, plugin, lava_job_info):
        asyncio_patch(
            mocker, "fc_server.plugins.lava.Plugin.lava_get_job_info", lava_job_info
        )
        ret = await plugin._Plugin__get_job_tags("0")
        assert ret == ("0", [])

        asyncio_patch(mocker, "fc_server.plugins.lava.Plugin.lava_get_job_info", None)

        ret = await plugin._Plugin__get_job_tags("0")
        assert ret is None

    @pytest.mark.asyncio
    async def test_get_device_tags(
        self, asyncio_patch, mocker, plugin, lava_device_info_unknown
    ):
        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.lava_get_device_info",
            lava_device_info_unknown,
        )
        ret = await plugin._Plugin__get_device_tags("$resource1")
        assert ret == ("$resource1", [])

        asyncio_patch(
            mocker, "fc_server.plugins.lava.Plugin.lava_get_device_info", None
        )
        ret = await plugin._Plugin__get_device_tags("$resource1")
        assert ret is None

    @pytest.mark.asyncio
    async def test_get_device_info(
        self,
        asyncio_patch,
        mocker,
        plugin,
        lava_device_info_unknown,
        lava_device_info_good,
    ):
        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.lava_get_device_info",
            lava_device_info_unknown,
        )
        ret = await plugin._Plugin__get_device_info("$resource1")
        assert ret == lava_device_info_unknown

        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.lava_get_device_info",
            lava_device_info_good,
        )
        ret = await plugin._Plugin__get_device_info("$resource1", clear=True)
        assert ret == lava_device_info_good

        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.lava_get_device_info",
            lava_device_info_unknown,
        )
        ret = await plugin._Plugin__get_device_info("$resource1")
        assert ret == lava_device_info_good

    @pytest.mark.asyncio
    async def test_force_kick_off(
        self, asyncio_patch, mocker, plugin, lava_device_info_with_job
    ):
        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.lava_get_device_info",
            lava_device_info_with_job,
        )

        mocker_lava_cancel_job = asyncio_patch(
            mocker, "fc_server.plugins.lava.Plugin.lava_cancel_job", MagicMock()
        )
        await plugin.force_kick_off("$resource1")
        mocker_lava_cancel_job.assert_called_with("1")

    @pytest.mark.asyncio
    async def test_seize_resource(self, asyncio_patch, mocker, coordinator, plugin):
        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin._Plugin__get_device_tags",
            ("$resource1", []),
        )

        mocker_coordinate_resources = asyncio_patch(
            mocker,
            "fc_server.core.coordinator.Coordinator.coordinate_resources",
            MagicMock(),
        )

        plugin.job_tags_cache["0"] = []

        await plugin._Plugin__seize_resource(coordinator, "0", ["$resource1"])
        mocker_coordinate_resources.assert_called()

    @pytest.mark.parametrize(
        "device, seize",
        [
            (
                [
                    {
                        "current_job": None,
                        "health": "Unknown",
                        "hostname": "$resource1",
                        "pipeline": True,
                        "state": "Idle",
                        "type": "docker",
                    }
                ],
                False,
            ),
            (
                [
                    {
                        "current_job": "1",
                        "health": "Unknown",
                        "hostname": "$resource1",
                        "pipeline": True,
                        "state": "Running",
                        "type": "docker",
                    }
                ],
                True,
            ),
        ],
    )
    @pytest.mark.asyncio
    async def test_schedule(
        self, asyncio_patch, mocker, plugin, coordinator, device, seize
    ):
        asyncio_patch(mocker, "fc_server.plugins.lava.Plugin.lava_get_devices", device)

        queued_job_info = [
            {
                "description": "foo",
                "id": "0",
                "requested_device_type": "docker",
                "submitter": "bar",
            }
        ]
        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.lava_get_queued_jobs",
            queued_job_info,
        )

        job_info = {
            "description": "foo",
            "device": None,
            "device_type": "docker",
            "end_time": None,
            "Health": "Unknown",
            "health_check": False,
            "id": "0",
            "pipeline": True,
            "start_time": None,
            "state": "Submitted",
            "submit_time": "20181101T07:13:06",
            "submitter": "foobar",
            "tags": [],
            "visibility": "Public",
        }
        asyncio_patch(
            mocker, "fc_server.plugins.lava.Plugin.lava_get_job_info", job_info
        )

        device_tags = ("$resource1", [])
        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin._Plugin__get_device_tags",
            device_tags,
        )

        async def mocker_coro():
            pass

        mocker.patch(
            "fc_server.plugins.lava.Plugin._Plugin__reset_possible_resource",
            return_value=mocker_coro(),
        )

        async def mocker_seize():
            pass

        mocker_seize = mocker.patch(
            "fc_server.plugins.lava.Plugin._Plugin__seize_resource",
            return_value=mocker_seize(),
        )

        await plugin.schedule(coordinator)
        assert mocker_seize.called == seize



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


import pytest

from fc_guarder.guarder import Guarder


class TestGuarder:
    @pytest.mark.parametrize("mode, code", [("good", 0), ("bad", 1)])
    def test_ping(self, mocker, mode, code):
        class Output:
            def __init__(self):
                self.status_code = None
                self.text = None

            def __call__(self, mode):
                if mode == "good":
                    self.status_code = 200
                    self.text = "pong"
                else:
                    self.status_code = 502
                    self.text = "error"
                return self

        mocker.patch(
            "requests.get",
            return_value=Output()(mode),
        )

        ret = Guarder.ping()
        assert ret == code



# -*- coding: utf-8 -*-
#
# Copyright 2023 NXP
#
# SPDX-License-Identifier: MIT


import asyncio
from unittest.mock import MagicMock

import pytest

from fc_server.plugins.labgrid import Plugin as LabgridPlugin
from fc_server.plugins.lava import Plugin as LavaPlugin


# pylint: disable=protected-access
@pytest.fixture(name="plugins")
def framework_plugins(coordinator):
    return coordinator._Coordinator__framework_plugins


@pytest.fixture(name="lava_plugin")
def get_lava_plugin(plugins):
    return plugins[0]


@pytest.fixture(name="labgrid_plugin")
def get_labgrid_plugin(plugins):
    return plugins[1]


# pylint: disable=protected-access
class TestCoordinator:
    def test_init_coordinator(self, lava_plugin, labgrid_plugin, coordinator):
        assert isinstance(lava_plugin, LavaPlugin)
        assert isinstance(labgrid_plugin, LabgridPlugin)

        assert coordinator._Coordinator__framework_priorities == {
            "lava": 1,
            "labgrid": 2,
        }
        assert coordinator._Coordinator__framework_seize_strategies == {
            "lava": True,
            "labgrid": False,
        }
        assert isinstance(
            coordinator._Coordinator__default_framework_plugin, LavaPlugin
        )

    @pytest.mark.asyncio
    async def test_init_frameworks(self, asyncio_patch, mocker, coordinator):
        ret = [
            {
                "current_job": "null",
                "health": "Unknown",
                "hostname": "docker-01",
                "pipeline": "true",
                "state": "Idle",
                "type": "docker",
            },
            {
                "current_job": "null",
                "health": "Unknown",
                "hostname": "docker-02",
                "pipeline": "true",
                "state": "Idle",
                "type": "docker",
            },
        ]
        asyncio_patch(mocker, "fc_server.plugins.lava.Plugin.lava_get_devices", ret)

        ret = "docker-01\ndocker-02"
        asyncio_patch(
            mocker, "fc_server.plugins.labgrid.Plugin.labgrid_get_places", ret
        )

        mocker_asyncio_gather = asyncio_patch(mocker, "asyncio.gather", MagicMock())

        await coordinator._Coordinator__init_frameworks()

        mocker_asyncio_gather.assert_called_with(*[])

    def test_is_default_framework(self, lava_plugin, labgrid_plugin, coordinator):
        assert coordinator.is_default_framework(lava_plugin)
        assert not coordinator.is_default_framework(labgrid_plugin)

    def test_get_low_priority_frameworks(self, coordinator):
        assert coordinator._Coordinator__get_low_priority_frameworks("lava") == [
            "labgrid"
        ]
        assert coordinator._Coordinator__get_low_priority_frameworks("labgrid") == []

    @pytest.mark.parametrize(
        "patch_ret, expected",
        [
            # disconnect successful from default framework, and maintenanced by fc
            (
                (True, True),
                (True, ["$resource1"], []),
            ),
            # disconnect successful from default framework, but maintenanced by user
            (
                (True, False),
                (True, [], []),
            ),
            # disconnect fail from default framework due to device busy or api failure,
            # but maintenanced by fc due to race condition
            (
                (False, True),
                (False, [], ["$resource1"]),
            ),
            # disconnect fail from default framework due to device busy or api failure,
            # and without maintenance
            (
                (False, False),
                (False, [], []),
            ),
        ],
    )
    @pytest.mark.asyncio
    async def test_is_resource_available(
        self,
        asyncio_patch,
        mocker,
        lava_plugin,
        labgrid_plugin,
        coordinator,
        patch_ret,
        expected,
    ):
        assert await coordinator.is_resource_available(lava_plugin, "$resource1")

        asyncio_patch(
            mocker,
            "fc_server.plugins.lava.Plugin.default_framework_disconnect",
            patch_ret,
        )
        assert (
            await coordinator.is_resource_available(labgrid_plugin, "$resource1")
            == expected[0]
        )
        assert (
            coordinator._Coordinator__managed_issue_disconnect_resources == expected[2]
        )

    def test_is_resource_non_available(self, coordinator):
        assert not coordinator.is_resource_non_available("$resource1")

    def test_is_seized_resource(self, lava_plugin, labgrid_plugin, coordinator):
        assert not coordinator.is_seized_resource(lava_plugin, "$resource1")
        assert not coordinator.is_seized_resource(labgrid_plugin, "$resource1")

    def test_is_seized_job(self, coordinator):
        assert not coordinator.is_seized_job(0)

    def test_accept_resource(self, coordinator, lava_plugin):
        coordinator.accept_resource("$resource1", lava_plugin)
        assert coordinator.managed_resources_status["$resource1"] == "lava"

    @pytest.mark.asyncio
    async def test_return_resource(self, coordinator, lava_plugin):
        coordinator.accept_resource("$resource1", lava_plugin)
        await coordinator.return_resource("$resource1")
        assert coordinator.managed_resources_status["$resource1"] == "fc"

    def test_retire_resource(self, coordinator):
        coordinator.retire_resource("$resource1")
        assert coordinator.managed_resources_status["$resource1"] == "retired"

    def test_reset_resource(self, coordinator):
        coordinator.reset_resource("$resource1")
        assert coordinator.managed_resources_status["$resource1"] == "fc"

    @pytest.mark.asyncio
    async def test_coordinate_resources_basic(self, coordinator, lava_plugin):
        assert await coordinator.coordinate_resources(lava_plugin, 0, *[]) == []

        assert await coordinator.coordinate_resources(lava_plugin, 0, "$resource1") == [
            "$resource1"
        ]

    @pytest.mark.asyncio
    async def test_coordinate_resources_seize_from_lava(
        self, asyncio_patch, mocker, coordinator, lava_plugin
    ):
        coordinator.accept_resource("$resource1", lava_plugin)
        mocker_force_kick_off = asyncio_patch(
            mocker, "fc_server.plugins.labgrid.Plugin.force_kick_off", MagicMock()
        )
        assert await coordinator.coordinate_resources(lava_plugin, 0, "$resource1") == [
            "$resource1"
        ]
        mocker_force_kick_off.assert_not_called()

    @pytest.mark.parametrize("seized_resource_awared", [True, False])
    @pytest.mark.asyncio
    async def test_coordinate_resources_seize_from_labgrid(
        self,
        asyncio_patch,
        mocker,
        coordinator,
        lava_plugin,
        labgrid_plugin,
        seized_resource_awared,
    ):
        coordinator.accept_resource("$resource1", labgrid_plugin)

        mocker_force_kick_off = asyncio_patch(
            mocker, "fc_server.plugins.labgrid.Plugin.force_kick_off", MagicMock()
        )

        mocker_reset_resource = asyncio_patch(
            mocker, "fc_server.core.coordinator.Coordinator.reset_resource", None
        )
        native_sleep = asyncio.sleep
        if not seized_resource_awared:
            asyncio_patch(mocker, "asyncio.sleep", MagicMock())

        assert await coordinator.coordinate_resources(lava_plugin, 0, "$resource1") == [
            "$resource1"
        ]

        if seized_resource_awared:
            coordinator.accept_resource("$resource1", get_lava_plugin)
        await native_sleep(0)

        mocker_force_kick_off.assert_called_with("$resource1")
        assert mocker_reset_resource.called == (not seized_resource_awared)



# -*- coding: utf-8 -*-
#
# Copyright 2024 NXP
#
# SPDX-License-Identifier: MIT


import pytest

from fc_server.core.config import Config


class TestConfig:
    @pytest.mark.usefixtures("remove_config_path")
    def test_no_config_file(self):
        with pytest.raises(SystemExit):
            Config.parse("")

    def test_normal(self):
        Config.parse("")

        assert Config.raw_managed_resources == {
            "$farm_type": {"$device_type": ["$resource1", "$resource2"]}
        }
        assert Config.registered_frameworks == ["lava", "labgrid"]
        assert Config.frameworks_config == {
            "lava": {
                "default": True,
                "friendly_status": "occupied(lava)",
                "identities": "$lava_identity",
                "priority": 1,
            },
            "labgrid": {
                "friendly_status": "occupied(labgrid)",
                "lg_crossbar": "ws://$labgrid_crossbar_ip:20408/ws",
                "priority": 2,
                "seize": False,
            },
            "fc": {
                "friendly_status": "idle",
            },
        }
        assert Config.priority_scheduler
        assert Config.api_server == {"port": 8600, "publish_port": 8600}
